/******************************************************************************
Description:

  A simple example CIF_Comm program to test and demonstrate the C++
  implementation.  This program demonstrates the base set of CIF_Comm classes
  and methods necessary to communicate data.  More advanced features can be
  found in the other example programs.

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/examples/CIF_Comm_Simple_Example.C,v $
  $Date: 1998/09/25 18:31:49 $
  $Revision: 1.5 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Toolkit.H"
#include "CIF_Comm.H"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/******************************************************************************
			     Function definitions
******************************************************************************/

/*
 * usage()
 */
void
usage(
    char *					exec_name)
{
    fprintf(stderr, "\nusage: %s [-c <nmsgs>] [-f] [-l] [-t <timeout>] "
	    "<-s|-r> <URL>\n\n", exec_name);
    fprintf(stderr, "    -c\tmessage count\n");
    fprintf(stderr, "    -f\trepeat forever\n");
    fprintf(stderr, "    -l\tlistener\n");
    fprintf(stderr, "    -r\treciever\n");
    fprintf(stderr, "    -s\tsender\n");
    fprintf(stderr, "    -t\ttimeout value\n\n");

    exit(1);
}


/*
 * main()
 */
int    
main(
    int					argc,
    char *				argv[])
{
    const int				MAX_MSG_SZ = 512;
    const int				ARGC_REQ = 1;

    int					argn;
    int					msg_cnt = 1;
    CIF_Bool				forever = false;
    CIF_Bool				listener = false;
    CIF_Bool				sender = false;
    CIF_Bool				receiver = false;
    int					timeout = -1;    
    char *				url;

    /*
     * Process arguments
     */
    argn = 0;
    while(++argn < argc - ARGC_REQ)
    {
	if (strcmp(argv[argn], "-c") == 0
	    && ++argn < argc - ARGC_REQ)
	{
	    msg_cnt = atoi(argv[argn]);
	    continue;
	}
	else if (strcmp(argv[argn], "-f") == 0)
	{
	    forever = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-l") == 0)
	{
	    listener = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-s") == 0)
	{
	    sender = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-r") == 0)
	{
	    receiver = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-t") == 0
		 && ++argn < argc - ARGC_REQ)
	{
	    timeout = atoi(argv[argn]);
	    continue;
	}
	else
	{
	    fprintf(stderr, "unknown argument: %s\n\n", argv[argn]);
	    usage(argv[0]);
	}
    }

    /*
     * Verify that all of the necessary arguments were specified
     */
    if (argn + ARGC_REQ == argc)
    {
	url = new char [strlen(argv[argn])+1];
	strcpy(url, argv[argn]);
    }
    else
    {
	usage(argv[0]);
    }

    if (!sender && !receiver)
    {
	usage(argv[0]);
    }
    

    /*
     * Activate CIF_Toolkit
     *
     * CIF_Toolkit::activate() and CIF_Comm::activate() must be executed before
     * any other CIF_Toolkit or CIF_Comm methods are called.  The activation
     * methods contain reference counting, so each module which uses one of the
     * CIF libraries can safely envoke the libraries activation method without
     * worrying about the potential problems associated with reactivation or
     * early deactivation.  However, it is important that all modules which
     * perform activation also perform deactivation when the have finished
     * using the CIF library.  Failing to perform deactivation can have
     * unpredicable results for some libraries.
     */
    CIF_Toolkit::activate();
    CIF_Comm::activate();

    /*
     * Establish a listener object and the wait for an incoming connection
     */
    CIF_Comm_Listener *			cif_listener;
    CIF_Bool				rc;

    if (listener)
    {
	printf("Establishing listener object for URL=%s\n", url);

	cif_listener = CIF_Comm_Factory::get_Listener(url);

	printf("get_Listener() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(cif_listener != CIF_NULL);

	printf("\nListening at %s\n\n", cif_listener->get_URL());
    }
    

    do
    {
	CIF_Comm_Connection *		cif_conn;

	if (listener)
	{
	    /*
	     * Wait for a remote process contact the listener and form a
	     * new connection.
	     */
	    printf("Waiting for connection\n");

	    if (timeout >= 0)
	    {
		cif_conn = cif_listener->get_Connection(timeout);
	    }
	    else
	    {
		cif_conn = cif_listener->get_Connection();
	    }
	    
	    printf("get_Connection() %s\n",
		   CIF_Comm::status.get_Description());
	    assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	    assert(cif_conn != CIF_NULL);
	}
	else
	{
	    /*
	     * Attempt to establish a new connection
	     */
	    printf("Establishing connection object for URL=%s\n", url);

	    if (timeout >= 0)
	    {
		cif_conn = CIF_Comm_Factory::get_Connection(url, timeout);
	    }
	    else
	    {
		cif_conn = CIF_Comm_Factory::get_Connection(url);
	    }

	    printf("get_Connection() %s\n",
		   CIF_Comm::status.get_Description());
	    assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	    assert(cif_conn != CIF_NULL);
	}
	printf("Connection established.\n");
	
	int				cnt;

	if (sender)
	{
	    /*
	     * Send msg_cnt number of messages
	     */
	    printf("Sending %d messages\n", msg_cnt);
	    
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		CIF_Byte		buf[MAX_MSG_SZ];
	    
		sprintf((char *) buf, "data for message %d", cnt);
		rc = cif_conn->send(buf, strlen((char *) buf) + 1, cnt);

		printf("send() %s\n",
		       CIF_Comm::status.get_Description());
		assert(rc == true);
		assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	    }
	}
	else
	{
	    /*
	     * Wait for the first message to be received.  This is not
	     * necessary since receive will block until a message is receive;
	     * however, it demostrates the use of the poll() method.
	     */
	    while (cif_conn->poll() == false)
	    {
		assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	    }

	    /*
	     * Revceive msg_cnt number of messages
	     */
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		CIF_Comm_Message		message;
	    
		rc = cif_conn->receive(message);

		printf("receive() %s\n",
		       CIF_Comm::status.get_Description());
		assert(rc == true);
		assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);

		printf("message received, size=%d, tag=%d, data=%s\n",
		       message.get_Info().size,
		       message.get_Info().tag,
		       message.get_Data());
	    
	    }
	}

	/*
	 * Close the connection and delete the connection object
	 */
	printf("Communication complete.  Closing connection.\n");

	rc = cif_conn->close();

	printf("close() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(rc == true);

	delete cif_conn;
    }
    while (forever);

    /*
     * Close the listener and delete the listener object
     */
    if (listener)
    {
	printf("Closing listener.\n");

	rc = cif_listener->close();

	printf("close() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(rc == true);

	delete cif_listener;
    }
    
    /*
     * Detactivate CIF
     *
     * CIF_Comm::deactivate() must be called when the application has finished
     * using CIF_Comm.  Normally, this is done right before the application
     * exits.
     */
    CIF_Comm::deactivate();
    CIF_Toolkit::deactivate();

    exit(0);
}
